/*
 * Copyright © 2001, 2002 Havoc Pennington
 * Copyright © 2002 Red Hat, Inc.
 * Copyright © 2002 Sun Microsystems
 * Copyright © 2003 Mariano Suarez-Alvarez
 * Copyright © 2008, 2010 Christian Persch
 *
 * Mate-terminal is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Mate-terminal is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <errno.h>
#include <locale.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <gio/gio.h>

#include <gdk/gdk.h>
#include <gdk/gdkx.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <semaphore.h>

#include "terminal-accels.h"
#include "terminal-app.h"
#include "terminal-debug.h"
#include "terminal-intl.h"
#include "terminal-options.h"
#include "terminal-util.h"

#define TERMINAL_FACTORY_UNIX_FOLDER "gatuno-terminal"
#define TERMINAL_FACTORY_SERVICE_NAME_PREFIX  "factory"

int factory_fd_lock = -1;
int factory_unix_fd = -1;
guint factory_unix_watch = 0;

static char *
ay_to_string (GVariant *variant,
              GError **error)
{
	gsize len;
	const char *data;

	data = g_variant_get_fixed_array (variant, &len, sizeof (char));
	if (len == 0)
		return NULL;

	/* Make sure there are no embedded NULs */
	if (memchr (data, '\0', len) != NULL)
	{
		g_set_error_literal (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
		                     "String is shorter than claimed");
		return NULL;
	}

	return g_strndup (data, len);
}

static char **
ay_to_strv (GVariant *variant,
            int *argc)
{
	GPtrArray *argv;
	const char *data, *nullbyte;
	gsize data_len;
	gssize len;

	data = g_variant_get_fixed_array (variant, &data_len, sizeof (char));
	if (data_len == 0 || data_len > G_MAXSSIZE)
	{
		if (argc != NULL)
			*argc = 0;
		return NULL;
	}

	argv = g_ptr_array_new ();

	len = data_len;
	do
	{
		gssize string_len;

		nullbyte = memchr (data, '\0', len);

		string_len = nullbyte ? (gssize) (nullbyte - data) : len;
		g_ptr_array_add (argv, g_strndup (data, string_len));

		len -= string_len + 1;
		data += string_len + 1;
	}
	while (len > 0);

	if (argc)
		*argc = argv->len;

	/* NULL terminate */
	g_ptr_array_add (argv, NULL);
	return (char **) g_ptr_array_free (argv, FALSE);
}

static GVariant *
string_to_ay (const char *string)
{
	gsize len;
	char *data;

	len = strlen (string);
	data = g_strndup (string, len);

	return g_variant_new_from_data (G_VARIANT_TYPE ("ay"), data, len, TRUE, g_free, data);
}

typedef struct
{
	char *factory_name;
	TerminalOptions *options;
	int exit_code;
	char **argv;
	int argc;
} OwnData;

gboolean unix_socket_dgram_recv (GIOChannel *source, GIOCondition condition, gpointer data) {
	char *message = NULL, byte;
	int message_len;
	int s;
	
	TerminalOptions *options = NULL;
	GVariant *v_wd, *v_display, *v_sid, *v_envv, *v_argv;
	char *working_directory = NULL, *display_name = NULL, *startup_id = NULL;
	int initial_workspace = -1;
	char **envv = NULL, **argv = NULL;
	int argc;
	GError *error = NULL;
	GVariant *parameters;
	
	s = g_io_channel_unix_get_fd (source);
	
	message_len = recvfrom (s, &byte, 1, MSG_TRUNC | MSG_PEEK, NULL, NULL);
	
	message = g_malloc (message_len);
	
	if (message == NULL) {
		return TRUE;
	}
	
	message_len = recvfrom (s, message, message_len, 0, NULL, NULL);
	if (message_len < 0) {
		g_free (message);
		
		return TRUE;
	}
	
	parameters = g_variant_new_from_data (G_VARIANT_TYPE ("(ayayayayiay)"), message, message_len, FALSE, NULL, NULL);
	
	g_variant_get (parameters, "(@ay@ay@ay@ayi@ay)",
	               &v_wd, &v_display, &v_sid, &v_envv, &initial_workspace, &v_argv);

	working_directory = ay_to_string (v_wd, &error);
	if (error)
		goto out;
	display_name = ay_to_string (v_display, &error);
	if (error)
		goto out;
	startup_id = ay_to_string (v_sid, &error);
	if (error)
		goto out;
	envv = ay_to_strv (v_envv, NULL);
	argv = ay_to_strv (v_argv, &argc);

	_terminal_debug_print (TERMINAL_DEBUG_FACTORY,
	                       "Factory invoked with working-dir='%s' display='%s' startup-id='%s'"
	                       "workspace='%d'\n",
	                       working_directory ? working_directory : "(null)",
	                       display_name ? display_name : "(null)",
	                       startup_id ? startup_id : "(null)",
	                       initial_workspace);

	options = terminal_options_parse (working_directory,
	                                  display_name,
	                                  startup_id,
	                                  envv,
	                                  TRUE,
	                                  TRUE,
	                                  &argc, &argv,
	                                  &error,
	                                  NULL);
	
	if (options != NULL) {
		options->initial_workspace = initial_workspace;

		terminal_app_handle_options (terminal_app_get (), options, FALSE /* no resume */, &error);
		terminal_options_free (options);
	}
out:
	g_variant_unref (v_wd);
	g_free (working_directory);
	g_variant_unref (v_display);
	g_free (display_name);
	g_variant_unref (v_sid);
	g_free (startup_id);
	g_variant_unref (v_envv);
	g_strfreev (envv);
	g_variant_unref (v_argv);
	g_strfreev (argv);

	if (error != NULL) {
		g_error_free (error);
	}
	
	g_free (message);
	return TRUE;
}


static void send_ask_open (const char *name, OwnData *data)
{
	GError *error = NULL;
	char **envv;
	int i;
	GVariantBuilder builder;
	GVariant *value;
	GString *string;
	char *s;
	gsize len;
	struct sockaddr_un unix_dest;
	int fd_u, ret;
	
	_terminal_debug_print (TERMINAL_DEBUG_FACTORY,
	                       "Sending message to %s on the unix socket\n", name);

	if (data->options == NULL)
	{
		/* Already handled */
		data->exit_code = EXIT_SUCCESS;
		return;
	}
	fd_u = socket (AF_UNIX, SOCK_DGRAM, 0);
	
	if (fd_u < 0) {
		data->exit_code = EXIT_FAILURE;
		return;
	}
	
	_terminal_debug_print (TERMINAL_DEBUG_FACTORY,
	                       "Forwarding arguments to existing instance\n");

	g_variant_builder_init (&builder, G_VARIANT_TYPE ("(ayayayayiay)"));
    
	g_variant_builder_add (&builder, "@ay", string_to_ay (data->options->default_working_dir));
	g_variant_builder_add (&builder, "@ay", string_to_ay (data->options->display_name));
	g_variant_builder_add (&builder, "@ay", string_to_ay (data->options->startup_id));

	string = g_string_new (NULL);
	envv = g_get_environ ();
	for (i = 0; envv[i]; ++i)
	{
		if (i > 0)
			g_string_append_c (string, '\0');

		g_string_append (string, envv[i]);
	}
	g_strfreev (envv);

	len = string->len;
	s = g_string_free (string, FALSE);
	g_variant_builder_add (&builder, "@ay",
	                       g_variant_new_from_data (G_VARIANT_TYPE ("ay"), s, len, TRUE, g_free, s));

	g_variant_builder_add (&builder, "@i", g_variant_new_int32 (data->options->initial_workspace));

	string = g_string_new (NULL);

	for (i = 0; i < data->argc; ++i)
	{
		if (i > 0)
			g_string_append_c (string, '\0');
		g_string_append (string, data->argv[i]);
	}

	len = string->len;
	s = g_string_free (string, FALSE);
	g_variant_builder_add (&builder, "@ay",
	                       g_variant_new_from_data (G_VARIANT_TYPE ("ay"), s, len, TRUE, g_free, s));
	
	GVariant *todo;
	todo = g_variant_builder_end (&builder);
	
	unix_dest.sun_family = AF_UNIX;
	g_snprintf (unix_dest.sun_path, sizeof (unix_dest.sun_path), "%s%s%s%s%s", g_get_user_runtime_dir (), G_DIR_SEPARATOR_S, TERMINAL_FACTORY_UNIX_FOLDER, G_DIR_SEPARATOR_S, name);
	
	g_free (subcadena);
	ret = sendto (fd_u, g_variant_get_data (todo), g_variant_get_size (todo), 0, (struct sockaddr *) &unix_dest, sizeof (unix_dest));
	if (ret < 0) {
		g_printerr ("Failed to forward arguments: \n");
		data->exit_code = EXIT_FAILURE;
	}
	
	close (fd_u);
	/*value = g_dbus_connection_call_sync (connection,
	                                     data->factory_name,
	                                     TERMINAL_FACTORY_SERVICE_PATH,
	                                     TERMINAL_FACTORY_INTERFACE_NAME,
	                                     "HandleArguments",
	                                     g_variant_builder_end (&builder),
	                                     G_VARIANT_TYPE ("()"),
	                                     G_DBUS_CALL_FLAGS_NONE,
	                                     -1,
	                                     NULL,
	                                     &error);*/
	g_variant_unref (todo);
	data->exit_code = EXIT_SUCCESS;

	terminal_options_free (data->options);
	data->options = NULL;
}

/* Settings storage works as follows:
 *   ~/.config/gatuno-terminal/config
 *   ~/.config/gatuno-terminal/profiles/Foo/
 *
 * It's somewhat tricky to manage the profiles/ dir since we need to track
 * the list of profiles, but GSettings doesn't have a concept of notifying that
 * a directory has appeared or disappeared.
 *
 * Session state is stored entirely in the RestartCommand command line.
 *
 * The number one rule: all stored information is EITHER per-session,
 * per-profile, or set from a command line option. THERE CAN BE NO
 * OVERLAP. The UI and implementation totally break if you overlap
 * these categories. See mate-terminal 1.x for why.
 *
 * Don't use this code as an example of how to use GSettings - it's hugely
 * overcomplicated due to the profiles stuff. Most apps should not
 * have to do scary things of this nature, and should not have
 * a profiles feature.
 *
 */

/* Copied from libcaja/caja-program-choosing.c; Needed in case
 * we have no DESKTOP_STARTUP_ID (with its accompanying timestamp).
 */
static Time
slowly_and_stupidly_obtain_timestamp (Display *xdisplay)
{
	Window xwindow;
	XEvent event;

	{
		XSetWindowAttributes attrs;
		Atom atom_name;
		Atom atom_type;
		const char *name;

		attrs.override_redirect = True;
		attrs.event_mask = PropertyChangeMask | StructureNotifyMask;

		xwindow =
		    XCreateWindow (xdisplay,
		                   RootWindow (xdisplay, 0),
		                   -100, -100, 1, 1,
		                   0,
		                   CopyFromParent,
		                   CopyFromParent,
		                   (Visual *)CopyFromParent,
		                   CWOverrideRedirect | CWEventMask,
		                   &attrs);

		atom_name = XInternAtom (xdisplay, "WM_NAME", TRUE);
		g_assert (atom_name != None);
		atom_type = XInternAtom (xdisplay, "STRING", TRUE);
		g_assert (atom_type != None);

		name = "Fake Window";
		XChangeProperty (xdisplay,
		                 xwindow, atom_name,
		                 atom_type,
		                 8, PropModeReplace, (unsigned char *)name, strlen (name));
	}

	XWindowEvent (xdisplay,
	              xwindow,
	              PropertyChangeMask,
	              &event);

	XDestroyWindow(xdisplay, xwindow);

	return event.xproperty.time;
}

static char *
get_factory_name_for_display (const char *display_name)
{
	GString *name;
	const char *p;
	int len;
	
	len = strlen (TERMINAL_FACTORY_SERVICE_NAME_PREFIX) + strlen (display_name) + 10 /* NUL */;

	name = g_string_sized_new (len);
	g_string_append (name, TERMINAL_FACTORY_SERVICE_NAME_PREFIX);

	for (p = display_name; *p; ++p)
	{
		if (g_ascii_isalnum (*p))
			g_string_append_c (name, *p);
		else
			g_string_append_c (name, '_');
	}

	_terminal_debug_print (TERMINAL_DEBUG_FACTORY,
	                       "Factory name is \"%s\"\n", name->str);

	return g_string_free (name, FALSE);
}

void create_factory_unix_socket_folder (void) {
	char path[8192];
	
	g_snprintf (path, sizeof (path), "%s%s%s", g_get_user_runtime_dir (), G_DIR_SEPARATOR_S, TERMINAL_FACTORY_UNIX_FOLDER);
	
	/* Crear el directorio de los sockets unix */
	g_mkdir_with_parents (path, 0700);
}

int get_lock_factory (char *factory_name) {
	char path[8192];
	struct flock fl;
	int fd;
	
	g_snprintf (path, sizeof (path), "%s%s%s%s%s.lock", g_get_user_runtime_dir (), G_DIR_SEPARATOR_S, TERMINAL_FACTORY_UNIX_FOLDER, G_DIR_SEPARATOR_S, factory_name);
	
	fd = open (path, O_RDWR | O_CREAT, 0600);
	
	fl.l_start = 0;
	fl.l_len = 0;
	fl.l_type = F_WRLCK;
	fl.l_whence = SEEK_SET;
	if (fcntl(fd, F_SETLK, &fl) < 0) {
		//fputs("Another instance of this program is running.\n", stderr);
		return -1;
	}
	
	factory_fd_lock = fd;
	return 0;
}

void release_lock_factory (void) {
	if (factory_fd_lock != -1) {
		close (factory_fd_lock);
		factory_fd_lock = -1;
	}
}

int do_bind_factory_unix_socket (char *factory_name) {
	int s, ret;
	struct sockaddr_un name;
	
	s = socket (AF_UNIX, SOCK_DGRAM, 0);
	
	if (s < 0) {
		return -1;
	}
	
	name.sun_family = AF_UNIX;
	g_snprintf (name.sun_path, sizeof (name.sun_path), "%s%s%s%s%s", g_get_user_runtime_dir (), G_DIR_SEPARATOR_S, TERMINAL_FACTORY_UNIX_FOLDER, G_DIR_SEPARATOR_S, factory_name);
	
	unlink (name.sun_path);
	
	ret = bind (s, (struct sockaddr *) &name, sizeof (name));
	if (ret < 0) {
		close (s);
		
		return -1;
	}
	
	return s;
}

void release_factory_unix_socket (char *factory_name) {
	struct sockaddr_un name;
	if (factory_unix_fd != -1) {
		g_source_remove (factory_unix_watch);
		factory_unix_watch = 0;
		
		close (factory_unix_fd);
		factory_unix_fd = -1;
		
		g_snprintf (name.sun_path, sizeof (name.sun_path), "%s%s%s%s%s", g_get_user_runtime_dir (), G_DIR_SEPARATOR_S, TERMINAL_FACTORY_UNIX_FOLDER, G_DIR_SEPARATOR_S, factory_name);
		unlink (name.sun_path);
	}
}

static int
get_initial_workspace (void)
{
  int ret = -1;
  GdkWindow *window;
  guchar *data = NULL;
  GdkAtom atom;
  GdkAtom cardinal_atom;

  window = gdk_get_default_root_window();

  atom = gdk_atom_intern_static_string ("_NET_CURRENT_DESKTOP");
  cardinal_atom = gdk_atom_intern_static_string ("CARDINAL");

  if (gdk_property_get (window, atom, cardinal_atom, 0, 8, FALSE, NULL, NULL, NULL, &data)) {
	  ret = *(int *)data;
	  g_free (data);
  }
  return ret;
}

int
main (int argc, char **argv)
{
	int i;
	char **argv_copy;
	int argc_copy;
	const char *startup_id, *display_name, *home_dir;
	GdkDisplay *display;
	TerminalOptions *options;
	GError *error = NULL;
	char *working_directory;
	int ret = EXIT_SUCCESS;

	setlocale (LC_ALL, "");

	bindtextdomain (GETTEXT_PACKAGE, TERM_LOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);

	_terminal_debug_init ();
	
	/* Make a NULL-terminated copy since we may need it later */
	argv_copy = g_new (char *, argc + 1);
	for (i = 0; i < argc; ++i)
		argv_copy [i] = argv [i];
	argv_copy [i] = NULL;
	argc_copy = argc;

	startup_id = g_getenv ("DESKTOP_STARTUP_ID");

	working_directory = g_get_current_dir ();

	gdk_set_allowed_backends ("x11");

	/* Now change directory to $HOME so we don't prevent unmounting, e.g. if the
	 * factory is started by caja-open-terminal. See bug #565328.
	 * On failure back to /.
	 */
	home_dir = g_get_home_dir ();
	if (home_dir == NULL || chdir (home_dir) < 0)
		if (chdir ("/") < 0)
			g_warning ("Could not change working directory.");

	options = terminal_options_parse (working_directory,
	                                  NULL,
	                                  startup_id,
	                                  NULL,
	                                  FALSE,
	                                  FALSE,
	                                  &argc, &argv,
	                                  &error,
	                                  gtk_get_option_group (TRUE),
	                                  NULL);

	g_free (working_directory);

	if (options == NULL)
	{
		g_printerr (_("Failed to parse arguments: %s\n"), error->message);
		g_error_free (error);
		exit (EXIT_FAILURE);
	}
	
	g_set_application_name (_("Terminal"));

	/* Unset the these env variables, so they doesn't end up
	 * in the factory's env and thus in the terminals' envs.
	 */
	g_unsetenv ("DESKTOP_STARTUP_ID");
	g_unsetenv ("GIO_LAUNCHED_DESKTOP_FILE_PID");
	g_unsetenv ("GIO_LAUNCHED_DESKTOP_FILE");

	display = gdk_display_get_default ();
	display_name = gdk_display_get_name (display);
	options->display_name = g_strdup (display_name);

	if (options->startup_id == NULL)
	{
		/* Create a fake one containing a timestamp that we can use */
		Time timestamp;

		timestamp = slowly_and_stupidly_obtain_timestamp (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()));

		options->startup_id = g_strdup_printf ("_TIME%lu", timestamp);
	}

	if (options->use_factory)
	{
		OwnData *data;
		guint owner_id;
		GError * error = NULL;
		data = g_new (OwnData, 1);
		data->factory_name = get_factory_name_for_display (display_name);
		data->options = options;
		data->exit_code = -1;
		data->argv = argv_copy;
		data->argc = argc_copy;
		GIOChannel *io_channel;
		
		gtk_init(&argc, &argv);
		options->initial_workspace = get_initial_workspace ();
		
		/* El proceso del factory funciona así:
		 * Si me pidieron usar el factory, intentar conseguir el candado.
		 * Si obtenemos el candado, poner el socket a la escucha.
		 * Si obtenemos el socket a la escucha, somos el factory. Lanzar terminal
		 * Si no consigo el socket a la escucha, error. Lanzar terminal como si no hubiera factory.
		 * Si no obtengo el candado, enviar mensaje a la otra terminal.
		 */
		create_factory_unix_socket_folder ();
		
		if (get_lock_factory (data->factory_name) == 0) {
			factory_unix_fd = do_bind_factory_unix_socket (data->factory_name);
			
			if (factory_unix_fd >= 0) {
				/* Soy la factory */
				_terminal_debug_print (TERMINAL_DEBUG_FACTORY, "Acquired the name %s on the unix socket\n", data->factory_name);
				/* Instalar un vigilador del socket unix */
				io_channel = g_io_channel_unix_new (factory_unix_fd);
				
				factory_unix_watch = g_io_add_watch (io_channel, G_IO_IN, unix_socket_dgram_recv, NULL);
			} else {
				/* ¿Tengo lock pero no pude conseguir el socket unix? Raro, soltar el lock, puesto que no puedo recibir las peticiones del factory */
				release_lock_factory ();
				
				/* Correr la terminal de a solas */
			}
			
			/* Correr la terminal */
			if (!terminal_app_handle_options (terminal_app_get (), data->options, TRUE /* do resume */, &error))
			{
				g_printerr ("Failed to handle options: %s\n", error->message);
				g_error_free (error);
				data->exit_code = EXIT_FAILURE;
			} else {
				terminal_options_free (data->options);
				data->options = NULL;
				
				gtk_main ();
			}
			
			release_factory_unix_socket (data->factory_name);
			release_lock_factory ();
		} else {
			/* Ya hay otra factory corriendo, enviar el mensaje */
			send_ask_open (data->factory_name, data);
		}
		
		ret = data->exit_code;

		g_free (data->factory_name);
		g_free (data);

	}
	else
	{

		terminal_app_handle_options (terminal_app_get (), options, TRUE /* allow resume */, &error);
		terminal_options_free (options);

		if (error == NULL)
		{
			gtk_main ();
		}
		else
		{
			g_printerr ("Error handling options: %s\n", error->message);
			g_error_free (error);
			ret = EXIT_FAILURE;
		}
	}

	terminal_app_shutdown ();

	g_free (argv_copy);

	return ret;
}
